#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <mqueue.h>
#include <gmp.h>
#include <semaphore.h>
#include <sys/mman.h>
#include <sys/types.h>

/* Encryption schemes */
typedef enum enc_scheme{
    DF96
  , DF98
  , DF02
  , MMH
} enc_scheme;

/* Encryption and decryption execution time calculators */
struct timeval start_time, end_time;
int total_usecs, sum_total_usecs;
unsigned int enc_sum=0, enc_avg=0, enc_count=0, enc_msg_split_id =0, enc_avg_sum=0;
unsigned int dec_sum=0, dec_avg=0, dec_count=0;

//Sets the encryption scheme in the shared memory.
void set_encryption_scheme(enc_scheme encryption_scheme /* Encryption scheme name */);


//Returns the encryption scheme from the shared memory.
enc_scheme get_encryption_scheme();

void encrypt(
  mpz_t msg                         /* Partial message to be encrypted */
  , char sender_msg_split_buff[]    /* Encrypted message */
  , int message_id                  /* Id of the message */
  , int message_split_id            /* Partial message id  */
  , int message_split_id_sum        /* Sum of partial message ids */
  , mpz_symmetric_key *mpz_keys_ptr /* Various encryption keys */


/***************************************************************************************
  Operation performed by the function df_add_mul_Ph_encryption()
    *Encrypts the message using Domingo-Ferrer's additive and multiplicative 
     privacy homomorphism.
    *Calculates the encryption time and writes to a file named 
     encryption_exec_time.log 
    *Places the encrypted value in a character buffer.
***************************************************************************************/
void df_add_mul_Ph_encryption(
  mpz_t msg                         /* Partial message to be encrypted */
  , char sender_msg_split_buff[]    /* Encrypted message */
  , int message_id                  /* Id of the message */
  , int message_split_id            /* Partial message id */
  , int message_split_id_sum        /* Sum of partial message ids */
  , mpz_symmetric_key *mpz_keys_ptr /* Various encryption keys */
  , FILE *encrypt_time_file_handle  /* File with the encryption time */
);

/***************************************************************************************
  Operation performed by the function df_new_Ph_encryption()
    *Encrypts the message using Domingo-Ferrer's new privacy homomorphism. 
    *Calculates the encryption time and writes to a file named 
     encryption_exec_time.log 
    *Places the encrypted value in a character buffer.
***************************************************************************************/
void df_new_Ph_encryption(
  mpz_t msg                         /* Partial message to be encrypted */
  , char sender_msg_split_buff[]    /* Encrypted message */
  , int message_id                  /* Id of the message */
  , int message_split_id            /* Partial message id */
  , int message_split_id_sum        /* Sum of partial message ids */
  , mpz_symmetric_key *mpz_keys_ptr /* Various encryption keys */
  , FILE *encrypt_time_file_handle  /* File with the encryption time */
);

/***************************************************************************************
  Operation performed by the function df_field_encryption()
    *Encrypts the message using Domingo-Ferrer's field privacy homomorphism.
    *Calculates the encryption time and writes to a file named 
     encryption_exec_time.log 
    *Places the encrypted value in a character buffer.
***************************************************************************************/
void df_field_encryption(
  mpz_t msg                         /* Partial message to be encrypted */
  , char sender_msg_split_buff[]    /* Encrypted message */
  , int message_id                  /* Id of the message */
  , int message_split_id            /* Partial message id */
  , int message_split_id_sum        /* Sum of partial message ids */
  , mpz_symmetric_key *mpz_keys_ptr /* Various encryption keys */
  , FILE *encrypt_time_file_handle  /* File with the encryption time */
);

/***************************************************************************************
  Operation performed by the function mmh_encryption()
    *Encrypts the message using mixed multiplicative homomorphism. 
    *Calculates the encryption time and writes to a file named 
     encryption_exec_time.log 
    *Places the encrypted value in a character buffer.
***************************************************************************************/
void mmh_encryption(
  mpz_t msg                         /* Partial message to be encrypted */
  , char sender_msg_split_buff[]    /* Encrypted message */
  , int message_id                  /* Id of the message */
  , int message_split_id            /* Partial message id */
  , int message_split_id_sum        /* Sum of partial message ids */
  , mpz_symmetric_key *mpz_keys_ptr /* Various encryption keys */
  , FILE *encrypt_time_file_handle  /* File with the encryption time */
);

/***************************************************************************************
  Operation performed by the function decrypt()
    *Calls the appropriate decryption function depending upon the value of 
     the encryption scheme.
***************************************************************************************/
void decrypt(
  mpz_t encryption_msg_split_array[][255] /* Messages to be decrypted */
  , mpz_t plaintext                       /* Decrypted plaintext */
  , int previous_counter                  /* Number of partial messages */
  , mpz_symmetric_key *mpz_keys_ptr       /* Various encryption keys*/
);

/***************************************************************************************
  Operation performed by the function df_add_mul_Ph_decryption()
    *Sums up all the partial messages to get the entire message.
    *The message is then decrypted by using Domingo-Ferrer's 
     additive and multiplicative privacy homomorphism. 
    *Calculates the decryption time and writes to a file named 
     decryption_exec_time.log 
***************************************************************************************/
void df_add_mul_Ph_decryption(
  mpz_t encryption_msg_split_array[][255] /* Messages to be decrypted */
  , mpz_t plaintext                       /* Decrypted plaintext */
  , int previous_counter                  /* Number of partial messages */
  , mpz_symmetric_key *mpz_keys_ptr       /* Various encryption keys */
  , FILE *decrypt_time_file_handle        /* File with the decryption time */
);

/***************************************************************************************
  Operation performed by the function df_new_Ph_decryption()
    *Sums up all the partial messages to get the entire message.
    *The message is then decrypted by using Domingo-Ferrer's 
     new privacy homomorphism. 
    *Calculates the decryption time and writes to a file named 
     decryption_exec_time.log 
***************************************************************************************/
void df_new_Ph_decryption(
  mpz_t encryption_msg_split_array[][255] /* Messages to be decrypted */
  , mpz_t plaintext                       /* Decrypted plaintext */
  , int previous_counter                  /* Number of partial messages */
  , mpz_symmetric_key *mpz_keys_ptr       /* Various encryption keys */
  , FILE *decrypt_time_file_handle        /* File with the decryption time */
);

/***************************************************************************************
  Operation performed by the function df_field_decryption()
    *Sums up all the partial messages to get the entire message.
    *The message is decrypted by using Domingo-Ferrer's field 
     privacy homomorphism. 
    *Calculates the decryption time and writes to a file named 
     decryption_exec_time.log 
***************************************************************************************/
void df_field_decryption(
  mpz_t encryption_msg_split_array[][255] /* Messages to be decrypted */
  , mpz_t plaintext                       /* Decrypted plaintext */
  , int previous_counter                  /* Number of partial messages */
  , mpz_symmetric_key *mpz_keys_ptr       /* Various encryption keys */
  , FILE *decrypt_time_file_handle        /* File with the decryption time */
);

/***************************************************************************************
  Operation performed by the function mmh_decryption()
    *Sums up all the partial messages to get the entire message.
    *The message is decrypted by using mixed multiplicative homomorphism. 
    *Calculates the decryption time and writes to a file named 
     decryption_exec_time.log 
***************************************************************************************/
void mmh_decryption(
  mpz_t encryption_msg_split_array[][255] /* Messages to be decrypted */
  , mpz_t plaintext                       /* Decrypted plaintext */
  , int previous_counter                  /* Number of partial messages */
  , mpz_symmetric_key *mpz_keys_ptr       /* Various encryption keys */
  , FILE *decrypt_time_file_handle        /* File with the decryption time */
);

