//***************************************************************************************
//*  Copyright: National ICT Australia,  2007, 2008					*
//*  Developed at the Networks and Pervasive Computing program				*
//*  Author(s): Athanassios Boulis, Dimosthenis Pediaditakis				*
//*  This file is distributed under the terms in the attached LICENSE file.		*
//*  If you do not find this file, copies can be found by writing to:			*
//*											*
//*      NICTA, Locked Bag 9013, Alexandria, NSW 1435, Australia			*
//*      Attention:  License Inquiry.							*
//*											*
//***************************************************************************************



#include <math.h>
#include "WChannel_Interference.h"

#define EV ev.disabled()?(ostream&)ev:ev

#define CASTALIA_DEBUG (!printDebugInfo)?(ostream&)DebugInfoWriter::getStream():DebugInfoWriter::getStream()

#define THRESHOLD_PROB 0.05
#define GOOD_LINK_PROB 0.98

#define IDEAL_MODULATION_THRESHOLD_SNR_DB 5.0
#define IDEAL_MODULATION_GOOD_LINK_SNR_DB 8.5


Define_Module(WChannel_Interference);


void WChannel_Interference::initialize() {

	readIniFileParameters();
	
	if(modulationType == MODULATION_FSK)
	{	
		thresholdSNRdB = 10.0 * log10(   (-2.0*(dataRate/noiseBandwidth)) * log( 2.0 * (1.0 -  pow(THRESHOLD_PROB, 1.0/(8.0*minumumPacketSize))) )   );
		goodLinkSNRdB =  10.0 * log10(   (-2.0*(dataRate/noiseBandwidth)) * log( 2.0 * (1.0 -  pow(GOOD_LINK_PROB, 1.0/(8.0*minumumPacketSize))) )   );
	}
	else
	if(modulationType == MODULATION_PSK)
	{
		thresholdSNRdB = 10.0 * log10(   (dataRate/noiseBandwidth) * pow(erfcInv(2.0 * ( 1.0 - pow(THRESHOLD_PROB, 1.0/(8.0*minumumPacketSize)) )), 2.0)   );
		goodLinkSNRdB = 10.0 * log10(   (dataRate/noiseBandwidth) * pow(erfcInv(2.0 * ( 1.0 - pow(GOOD_LINK_PROB, 1.0/(8.0*minumumPacketSize)) )), 2.0)   );
	}
	else //(modulationType == MODULATION_IDEAL)
	{
		thresholdSNRdB = IDEAL_MODULATION_THRESHOLD_SNR_DB;
		goodLinkSNRdB = IDEAL_MODULATION_GOOD_LINK_SNR_DB;
	}
	
	
	
	// activeNodes is empty
	activeNodes = NULL;
	
	//carrierSensingNodes is empty
	carrierSensingNodes = NULL;

	// Allocate and Initialize rxSignal
	rxSignal = new double** [numNodes];
	if (rxSignal==NULL) opp_error("Could not allocate array rxSignal\n");
	for (int i = 0; i < numNodes; i++) {
		rxSignal[i] = new double* [numNodes];
		if (rxSignal[i]==NULL) opp_error("Could not allocate array rxSignal[%d]\n", i);
		for (int j = 0; j < numNodes; j++){
			rxSignal[i][j] = new double [numTxPowerLevels];
			if (rxSignal[i][j]==NULL) opp_error("Could not allocate array rxSignal[%d][%d]\n", i, j);
		}
	}


	cTopology *topo = new cTopology("topo");
	topo->extractByModuleType("Node", NULL);


	double x1, x2, y1, y2, dist;
	double PLd; // path loss at distance dist, in dB

	for (int i = 0; i < numNodes; i++)
	{
		x1 = topo->node(i)->module()->par("xCoor");
		y1 = topo->node(i)->module()->par("yCoor");	

	    // For visualization and other debugging purposes
	    CASTALIA_DEBUG << "\n[Wireless Channel]: Location of Node[" << i << "]: " << x1 << ", " << y1;
		
	    for (int j = 0; j < numNodes; j++)
	    {
	    	if (i == j)
			{
				for (int k = 0; k < numTxPowerLevels; k++)
					rxSignal[i][j][k] = 100.0;  // something very big, so you cannot receive anything while transmitting
			}
			else
			{
				x2 = topo->node(j)->module()->par("xCoor");
				y2 = topo->node(j)->module()->par("yCoor");

				dist = sqrt((x2 - x1)*(x2 - x1) + (y2 - y1)*(y2 - y1));

				PLd = PLd0 + 10.0 * pathLossExponent * log10(dist/d0) + normal(0, sigma);

				// the Path Loss, **including the random part** remains
				// the same for different tx power levels, if we are
				// referring to the same receiver-transmitter pair.

				for (int k = 0; k < numTxPowerLevels; k++) 
					rxSignal[i][j][k] = txPowerLevels[k] - PLd;
			}
	    }
	}
	
	
	if(allBidirectionalLinks)
	{
		double snrA, snrB;
		double rndVar;
		
		for (int i = 0; i < numNodes; i++)
		{
			for (int j = 0; j < numNodes; j++)
		    {
		    	if(i < j)
				{
			    	for (int k = 0; k < numTxPowerLevels; k++) 
					{
						snrA = (rxSignal[i][j][k] - noiseFloor);
						snrB = (rxSignal[j][i][k] - noiseFloor);

						if( (snrA > goodLinkSNRdB) || (snrB > goodLinkSNRdB) )
						{
							if(snrA > snrB)
								rxSignal[j][i][k] = rxSignal[i][j][k];
							else
								rxSignal[i][j][k] = rxSignal[j][i][k]; 
						}
						else
						{
							rndVar = genk_dblrand(0);
										    					    		
				    		if(rndVar < 0.5)
				    			rxSignal[j][i][k] = rxSignal[i][j][k];
				    		else
				    			rxSignal[i][j][k] = rxSignal[j][i][k];
						}
					}
				}
		    }
		}
	}
	
	
	delete(topo);
	
	// If direct assignment of link qualities are given at the omnetpp.ini file
	// (either as RxPower or PRR) we parse the input and update rxSignal
	parseRxPowerMap();
	parsePrrMap();

  // For visualization purposes, draws a no-collission connectivity map
  if (printDebugInfo) {
    
    double SNR_dB, prob;
    int  packetByteSize = 25; // to determine the PRR we need to know the packet size
                              // we define a nominal min size of 25. It would be better
                              // if this was dynamically determined for every packet.
                              // We could also take into account interference. 
    
    for (int i = 0; i < numNodes; i++)
    {
	      CASTALIA_DEBUG << "\n[Wireless Channel]: Node " << i << " TX, Possible Receptions (node->prob): ";
	      for (int j = 0; j < numNodes; j++)
	      {
				if (j==i)
					continue; 
				
				SNR_dB = rxSignal[i][j][defaultPowerLevel] - noiseFloor;
						
				if ((SNR_dB < thresholdSNRdB) || (rxSignal[i][j][defaultPowerLevel] < receiverSensitivity))
					continue;
				else
				{
					// Calculate the prob of packet reception
					if(modulationType == MODULATION_FSK)
						prob = pow( 1.0 - 0.5 * exp( ((-0.5) * noiseBandwidth / dataRate)* pow(10.0, (SNR_dB/10.0))), (8.0*packetByteSize) );   //  1/1.28 = 0.781
					else
					if(modulationType == MODULATION_PSK)
					  	prob = pow( 1.0 - 0.5 * erfc( sqrt(pow(10.0,(SNR_dB/10.0)) * noiseBandwidth / dataRate)  ),  (8.0*packetByteSize));
					else //(modulationType == MODULATION_IDEAL)
						prob = 1.0;
					
					
					CASTALIA_DEBUG << j << "->" << prob << ", "; 
				}
	      }
    }
  
    CASTALIA_DEBUG << "\n\n\n";
  }  // end visualization debugging output

/*  Print-out the rxSignal[][][] table, for the defaultPowerLevel 
// printRxSignalTable();
printf("\n=========================================\n");

for (int i = 0; i < numNodes; i++)
      printf("[%i] \t\t", i);
      
 for (int i = 0; i < numNodes; i++) {
      printf("\n");
      
      for (int j = 0; j < numNodes; j++) {
      	printf("[%.1f] \t", rxSignal[i][j][defaultPowerLevel]);
      }
 }
 
 printf("\n=========================================\n");
*/

	// Allocate and Initialize maxInterferenceForCurrentTx ONLY if it is going
	// to be used
	if(collisionModel != NO_INTERFERENCE_NO_COLLISIONS)
	{
		maxInterferenceForCurrentTx = new double* [numNodes];
		if (maxInterferenceForCurrentTx==NULL) opp_error("Could not allocate array maxInterferenceForCurrentTx\n");
		for (int i = 0; i < numNodes; i++) {
			maxInterferenceForCurrentTx[i] = new double [numNodes];
			if (maxInterferenceForCurrentTx[i]==NULL) opp_error("Could not allocate array maxInterferenceForCurrentTx[%d]\n", i);
		}
		for ( int i = 0; i < numNodes; i++)
			for (int j = 0; j < numNodes; j++)
				maxInterferenceForCurrentTx[i][j] = -200.0; // something very small, approaching 0 mW
	}
	


	// Allocate and Initialize stats
	stats = new int* [numNodes];
	if (stats==NULL) opp_error("Could not allocate array stats\n");
	for (int i = 0; i < numNodes; i++) {
		stats[i] = new int [7];
		if (stats[i]==NULL) opp_error("Could not allocate array stats[%d]\n", i);
	}
	for ( int i = 0; i < numNodes; i++)
		for (int j = 0; j < 7; j++)
			stats[i][j] = 0;


	// Schedule a self message at the sim-time-limit so the simulation does not
	// finish before the sim-time-limit
	scheduleAt(ev.config()->getAsTime("General", "sim-time-limit"), new WChannel_GenericMessage("sim-time-limit message", WC_SIMTIME_LIMIT));
}


void WChannel_Interference::handleMessage(cMessage *msg) {
	//process the messages received
	WChannel_GenericMessage* message = dynamic_cast<WChannel_GenericMessage*>(msg);

	ListElement *node, *node2;
	int srcAddr = message->getSrcAddress();
	int packetByteSize = message->byteLength();
	//int destAddr = message->getDestAddress(); 	// this is not used for now
	int powerLevel, powerLevel2, srcPowerLevel;
	int tempAddr;
	double currentInterferencePower, SNR_dB, prob;
 
	switch (message -> kind()) {

		case WC_SIMTIME_LIMIT:
		{
			CASTALIA_DEBUG << "\n[Wireless Channel]: Received self message to ensure sim-time-limit is reached @time:" << simTime();
			
			break;
		}

		case WC_CARRIER_SENSE_BEGIN:
		{	node = new ListElement;
			node->index = srcAddr;
			carrierSensingNodes = addToFront(carrierSensingNodes, node);

			// Check the current carrier signal strength for this node
			// We adopt a simple testing: If any of the active nodes reaches
			// the carrier sensing node with power that exceeds the receiver 
			// sensitivity, we assume carrier is sensed. 
			// We do not add powers from different nodes.

			for (node = activeNodes; node != NULL; node = node -> next) { //iterate through the active list
				tempAddr = node->index;
				powerLevel = node->powerLevel;
				if ((powerLevel < 0) || (powerLevel >= numTxPowerLevels)) powerLevel = 0;  //assign the default value

				if (rxSignal[tempAddr][srcAddr][powerLevel] > receiverSensitivity) {
					// carrier sensed! send message to node.
					// Delay it by 0.1ms (the defined minimum carrier sense interval)
					sendDelayed( new cMessage("carrier signal sensed", WC_CARRIER_SENSED), 0.0001, "toNode", srcAddr);
					// no need to continue with the loop
					break;
				}
			}
			break;
		}


		case WC_CARRIER_SENSE_END:
		{	// just delete this node from the list of carrierSensingNodes
			carrierSensingNodes = deleteListElement(carrierSensingNodes, srcAddr);
			break;
		}



		case WC_PKT_BEGIN_TX:
		{			
			node = new ListElement;
			node->index = srcAddr;
			node->powerLevel = message->getPowerLevel();
			// add the transmitting node to the active list
			activeNodes = addToFront(activeNodes, node);
			// update the stats
			stats[srcAddr][0]++;

			// check whether carrier signal is sensed in the carrierSensingNodes due to this transmission
			// again we take a simple approach, we do not keep power seen before and add it.
			powerLevel = node->powerLevel;
			if ((powerLevel < 0) || (powerLevel >= numTxPowerLevels)) powerLevel = defaultPowerLevel;  //assign the default value

			for (node = carrierSensingNodes; node != NULL; node = node -> next) { //iterate through the active list
				tempAddr = node->index;

				if (rxSignal[srcAddr][tempAddr][powerLevel] > receiverSensitivity) {
				//if (rxSignal[srcAddr][tempAddr][powerLevel] - noiseFloor > thresholdSNRdB) {
					// carrier sensed! send message to node.
					// Delay it by 0.1ms (the defined minimum carrier sense interval)
					sendDelayed( new cMessage("carrier signal sensed", WC_CARRIER_SENSED), 0.0001, "toNode", tempAddr);
				}
			}
			break;
		}
		

		case WC_PKT_END_TX:
		{
			switch(collisionModel)
			{
				case NO_INTERFERENCE_NO_COLLISIONS:
				{
					
					for (node = activeNodes; node != NULL; node = node -> next) 
					{   
						tempAddr = node->index;
						
						if (tempAddr == srcAddr) 
						{
							powerLevel = node->powerLevel;
						
							if ((powerLevel < 0) || (powerLevel >= numTxPowerLevels)) 
								powerLevel = defaultPowerLevel;  //assign the default value
							
							srcPowerLevel = powerLevel; // set srcPowerLevel to be used later
						}
					}
					
					
					// print debug info about possible receptions if needed 
					if (printDebugInfo){
						
						// we need to copy the message because decapsulation affects it
						WChannel_GenericMessage* temp = (WChannel_GenericMessage *) message->dup();
						MAC_GenericFrame* macFrame = (check_and_cast<MAC_GenericFrame *>(temp->decapsulate()));
					
						if (macFrame->getHeader().frameType == MAC_PROTO_BEACON_FRAME) 
						CASTALIA_DEBUG <<  "\n[WChannel]: t= " << simTime() << " Node " << srcAddr << " TX beacon, Possible Receptions (node->prob): ";
						else
						CASTALIA_DEBUG <<  "\n[WChannel]: t= " << simTime() << " Node " << srcAddr << " TX, Possible Receptions (node->prob): ";
						
						delete temp;
						delete macFrame;
					}
					
					
					// It's the end of transmission from node srcAddr so we have to check based on the
					// rxSignal and the maxInterference experienced, which receivers get the transmission.
					for (int i = 0; i < numNodes; i++) {
		
						if (i == srcAddr) continue;  // a node cannot receive its own transmission
		
						if (rxSignal[srcAddr][i][srcPowerLevel] > receiverSensitivity)
							stats[srcAddr][1]++;	// update stats: possible reception
						else 
							continue;				// if not possible reception, go to the next node (next i)
		

	        			SNR_dB = rxSignal[srcAddr][i][srcPowerLevel] - noiseFloor;
        			
//printf("\n[%i]-->[%i]  rxSignal=%f dBm,  SNRdb=%f\n", srcAddr, i, rxSignal[srcAddr][i][srcPowerLevel], SNR_dB);

						if (SNR_dB > thresholdSNRdB)
						{
							// transmission has a good chance of being received, i.e. :
							// noise was not strong enough for packet to
							// be received at receiver i with prob < thresholdProb
								
								
							// Calculate the prob of packet reception
							if(modulationType == MODULATION_FSK)
								prob = pow( 1.0 - 0.5 * exp( ((-0.5) * noiseBandwidth / dataRate)* pow(10.0, (SNR_dB/10.0))), (8.0*packetByteSize) );   //  1/1.28 = 0.781
							else
							if(modulationType == MODULATION_PSK)
							  	prob = pow( 1.0 - 0.5 * erfc( sqrt(pow(10.0,(SNR_dB/10.0)) * noiseBandwidth / dataRate)  ),  (8.0*packetByteSize));
							else //(modulationType == MODULATION_IDEAL)
								prob = 1.0;

							
							// print debug info about possible receptions if needed
		          			CASTALIA_DEBUG << i << "->" << prob << ", ";

//printf("[%i]-->[%i]  P[Rx]=%f\n", srcAddr, i, prob);
		
							if ( (prob == 1.0) || ( genk_dblrand(1) <= prob ) ) 
							{  
								// notice that we are using generator 1 not 0						
								// Send it! we cannot send msg multiple times, we have to create copies
								cMessage *copy = (cMessage *) msg->dup();
								WChannel_GenericMessage *wcMsg = check_and_cast<WChannel_GenericMessage *>(copy);
								wcMsg->setRssi(rxSignal[srcAddr][i][srcPowerLevel]);
								send(wcMsg, "toNode", i);
								// update the stats, actual reception
								stats[srcAddr][4]++;
								stats[i][5]++;
							}
							else
							{
								stats[srcAddr][3]++;  // the transmission was not successful due to wireless channel errors
							}
							
						}
						else 
						{
							stats[srcAddr][2]++;  // the transmission was interferred
							stats[i][6]++;
						}
					}
		
					// delete srcAddr from the active nodes
					activeNodes = deleteListElement(activeNodes, srcAddr);
					
					
					break;
				}
				
				
				
				case ADDITIVE_INTERFERENCE_MODEL: case SIMPLE_COLLISION_MODEL:
				{
					// we calculate how much is the current interference to every transmission
					// from all the other transmissions (at the possible/eligible receivers)
					// and compare it with the max seen so far. This way we will know the max
					// experienced interference for each transmission when it ends and decide
					// on the reception packet prob (for each of the eligible receivers)
		
					for (node = activeNodes; node != NULL; node = node -> next) 
					{   //iterate through the active list
						tempAddr = node->index;
						
						powerLevel = node->powerLevel;
						
						if ((powerLevel < 0) || (powerLevel >= numTxPowerLevels)) 
							powerLevel = defaultPowerLevel;  //assign the default value
							
						if (tempAddr == srcAddr) 
							srcPowerLevel = powerLevel; // set srcPowerLevel to be used later
		
						
						for (int i = 0; i < numNodes; i++)
						{
							
							if (rxSignal[tempAddr][i][powerLevel] - noiseFloor > thresholdSNRdB) 
							{
								// got one potential receiver with index i
								// iterate once more through the active list to find the
								// interference to the transmision from node tempAddr "seen" at node i
								currentInterferencePower = -200.0; 	// a very small value in dBm
								for (node2 = activeNodes; node2 != NULL; node2 = node2 -> next)
								{
									// a node does not interfere with its own transmission
									if (node2 == node) continue;
									
									powerLevel2 = node2->powerLevel;
									
									if ((powerLevel2 < 0) || (powerLevel2 >= numTxPowerLevels)) 
										powerLevel2 = defaultPowerLevel;
									
									currentInterferencePower = addPower_dBm(rxSignal[node2->index][i][powerLevel2], currentInterferencePower);
								}//for_node2
								
								if (maxInterferenceForCurrentTx[tempAddr][i] < currentInterferencePower)
									maxInterferenceForCurrentTx[tempAddr][i] = currentInterferencePower;
							}
							
						}//for_i
					}
					
					
					
					// print debug info about possible receptions if needed 
					if (printDebugInfo){
						
						// we need to copy the message because decapsulation affects it
						WChannel_GenericMessage* temp = (WChannel_GenericMessage *) message->dup();
						MAC_GenericFrame* macFrame = (check_and_cast<MAC_GenericFrame *>(temp->decapsulate()));
					
						if (macFrame->getHeader().frameType == MAC_PROTO_BEACON_FRAME) 
						CASTALIA_DEBUG <<  "\n[WChannel]: t= " << simTime() << " Node " << srcAddr << " TX beacon, Possible Receptions (node->prob): ";
						else
						CASTALIA_DEBUG <<  "\n[WChannel]: t= " << simTime() << " Node " << srcAddr << " TX, Possible Receptions (node->prob): ";
						
						delete temp;
						delete macFrame;
					}
					
					
					
					// It's the end of transmission from node srcAddr so we have to check based on the
					// rxSignal and the maxInterference experienced, which receivers get the transmission.
					for (int i = 0; i < numNodes; i++) {
		
						if (i == srcAddr) continue;  // a node cannot receive its own transmission
		
						if (rxSignal[srcAddr][i][srcPowerLevel] > receiverSensitivity)
							stats[srcAddr][1]++;	// update stats: possible reception
						else 
							continue;				// if not possible reception, go to the next node (next i)
							
							
						if(collisionModel == SIMPLE_COLLISION_MODEL)
						{
							if( (maxInterferenceForCurrentTx[srcAddr][i] > noiseFloor) &&
							    (maxInterferenceForCurrentTx[srcAddr][i] > receiverSensitivity) )
							    SNR_dB = -1;
						}
						else //(collisionModel == ADDITIVE_INTERFERENCE_MODEL)
						{
							SNR_dB = rxSignal[srcAddr][i][srcPowerLevel] - addPower_dBm(noiseFloor, maxInterferenceForCurrentTx[srcAddr][i]);
						}
						
						if (SNR_dB > thresholdSNRdB)
						{	
							// transmission has a good chance of being received, i.e. :
							// noise + max interference were not strong enough for packet to
							// be received at receiver i with prob < thresholdProb
							
							
							// Calculate the prob of packet reception
							if(modulationType == MODULATION_FSK)
								prob = pow( 1.0 - 0.5 * exp( ((-0.5) * noiseBandwidth / dataRate)* pow(10.0, (SNR_dB/10.0))), (8.0*packetByteSize) );   //  1/1.28 = 0.781
							else
							if(modulationType == MODULATION_PSK)
							  	prob = pow( 1.0 - 0.5 * erfc( sqrt(pow(10.0,(SNR_dB/10.0)) * noiseBandwidth / dataRate)  ),  (8.0*packetByteSize));
							else //(modulationType == MODULATION_IDEAL)
								prob = 1.0;
								
														
							// print debug info about possible receptions if needed
		          			CASTALIA_DEBUG << i << "->" << prob << ", "; 
		
		
							if ( (prob == 1.0) || ( genk_dblrand(1) <= prob ) ) 
							{
								// notice that we are using generator 1 not 0						
								// Send it! we cannot send msg multiple times, we have to create copies
								cMessage *copy = (cMessage *) msg->dup();
								WChannel_GenericMessage *wcMsg = check_and_cast<WChannel_GenericMessage *>(copy);
								wcMsg->setRssi(rxSignal[srcAddr][i][srcPowerLevel]);
								send(wcMsg, "toNode", i);
								// update the stats, actual reception
								stats[srcAddr][4]++;
								stats[i][5]++;
							}
							else
							{
								stats[srcAddr][3]++;  // the transmission was not successful due to wireless channel errors
							}
							
						}
						else 
						{
							stats[srcAddr][2]++;  // the transmission was interferred
							stats[i][6]++;
						}
					}
		
					// delete srcAddr from the active nodes and minimize the row in maxInterferenceForCurrentTx
					activeNodes = deleteListElement(activeNodes, srcAddr);
					for (int i = 0; i < numNodes; i++)
						maxInterferenceForCurrentTx[srcAddr][i] = -200.0; // a very small value in dBm
					
					
					
					break;
				}
				
				
				
				
				case COMPLEX_INTERFERENCE_MODEL:
				{
					opp_error("The specified \"Collision model\" in the omnetpp.ini file is not yet supported/implemented.");
					break;
				}
				
				
				
				
				default : 
				{
					opp_error("Error in specifying \"Collision model\" in the omnetpp.ini.");
					break;
				}
			}//switch(collisionModel)

			break;
		}//case WC_PKT_END_TX
		

		default:
		{
			CASTALIA_DEBUG << "\n[WChannel]:\n Warning: Wireless Channel received unknown message!";
			break;
		}
	}

	delete msg;
}


void WChannel_Interference::finish() {

	if(printStatistics)
	{
		// output aggregate transmission/reception info
		// useful for debugging purposes
		for(int i = 0;  i < numNodes; i++)
		{
			CASTALIA_DEBUG << "\n\n## Wireless Channel --> Node[" << i << "] ##";
			CASTALIA_DEBUG << "\n - transmissions = " << stats[i][0];
			CASTALIA_DEBUG << "\n     These transmissions resulted in " << stats[i][1] << " possible receptions at other nodes.";
			CASTALIA_DEBUG << "\n     From these possible receptions:";
			CASTALIA_DEBUG << "\n            " << stats[i][2] << " were interferred, ";
			CASTALIA_DEBUG << "\n            " << stats[i][3] << " were failed due to Wireless Channel errors, ";
			CASTALIA_DEBUG << "\n            " << stats[i][4] << " were strong enough to be received (radio and MAC states allowing).";
			CASTALIA_DEBUG << "\n - receptions = " << stats[i][5] << " (packets strong enough to be received (radio and MAC states allowing)";
			CASTALIA_DEBUG << "\n - collisions = " << stats[i][6] << " (packets intented for Node[" << i << "] that got interferred)\n";
		}
	}
	
	
	// Delete dynamically allocated arrays
	// delete rxSignal
	for (int i = 0; i < numNodes; i++){
		for (int j = 0; j < numNodes; j++)
			delete[] rxSignal[i][j];			// the delete[] operator releases memory allocated with new []

		delete[] rxSignal[i];
	}
	delete[] rxSignal;


	if(collisionModel != NO_INTERFERENCE_NO_COLLISIONS)
	{
		//delete maxInterferenceForCurrentTx
		for (int i = 0; i < numNodes; i++)
			delete[] maxInterferenceForCurrentTx[i];
		delete[] maxInterferenceForCurrentTx;
	}
	
	
	//delete stats
	for (int i = 0; i < numNodes; i++)
		delete[] stats[i];
	delete[] stats;

	// delete activeNodes elements
	for (ListElement *el = activeNodes; el != NULL; ) {
		activeNodes = el->next;
		delete el;
		el = activeNodes;
	}

	// delete carrierSensingNodes elements
	for (ListElement *el = carrierSensingNodes; el != NULL; ) {
		carrierSensingNodes = el->next;
		delete el;
		el = carrierSensingNodes;
	}
	
	//close the output stream that CASTALIA_DEBUG is writing to
	DebugInfoWriter::closeStream();
}



void WChannel_Interference::readIniFileParameters(void)
{
	string tempDebugFname(parentModule()->par("debugInfoFilename"));
	DebugInfoWriter::setDebugFileName(tempDebugFname);
	
	printDebugInfo = par("printDebugInfo");
	
	printStatistics = par("printStatistics");

	numNodes = parentModule()->par("numNodes");
	
	pathLossExponent = par("pathLossExponent");
	
	sigma = par("sigma");
	
	PLd0 = par("PLd0");
	
	d0 = par("d0");
	
	rxSignal_ConnectivityMap = par("rxSignal_ConnectivityMap");
	
	PRR_ConnectivityMap = par("PRR_ConnectivityMap");
	
	//find the minimum packet size
	cModule *node_0;
	if(parentModule()->findSubmodule("node", 0) != -1)
		node_0 = check_and_cast<cModule*>(parentModule()->submodule("node", 0));
	else
		opp_error("\n[Wireless Channel]:\n Error in geting a valid reference to  node[0].\n");
		
	if( node_0->findSubmodule("nodeApplication")  &&
		node_0->findSubmodule("networkInterface") &&
		node_0->submodule("networkInterface")->findSubmodule("MAC") &&
		node_0->submodule("networkInterface")->findSubmodule("Radio") )
	{
		int appOverhead = node_0->submodule("nodeApplication")->par("packetHeaderOverhead");
		int macOverhead = node_0->submodule("networkInterface")->submodule("MAC")->par("macFrameOverhead");
		int radioOverhead = node_0->submodule("networkInterface")->submodule("Radio")->par("phyFrameOverhead");
		minumumPacketSize = (double)(appOverhead + macOverhead + radioOverhead + 2);
		
		
		dataRate = node_0->submodule("networkInterface")->submodule("Radio")->par("dataRate");
		
		receiverSensitivity = node_0->submodule("networkInterface")->submodule("Radio")->par("receiverSensitivity");
		
		noiseBandwidth = node_0->submodule("networkInterface")->submodule("Radio")->par("noiseBandwidth");
		
		noiseFloor = node_0->submodule("networkInterface")->submodule("Radio")->par("noiseFloor");
		

		// 0 --> MODULATION_IDEAL (no transmission errors)
		// 1 --> FSK (noncoherent) -- for MICA2 motes
		// 2 --> BPSK, QPSK, OQPSK -- for CC2420 (Telos motes)
		modulationType = node_0->submodule("networkInterface")->submodule("Radio")->par("modulationType");
		if(modulationType == 0)
			modulationType = MODULATION_IDEAL;
		else
		if(modulationType == 1)
			modulationType = MODULATION_FSK;
		else
		if(modulationType == 2)
			modulationType = MODULATION_PSK;
		else
			opp_error("\n[Wireless Channel]: Illegal value of parameter \"modulationType\" in omnetpp.ini\n" );
		
		
		
		encodingType = node_0->submodule("networkInterface")->submodule("Radio")->par("encodingType");
		switch(encodingType)
		{
			case 0:
			{
				encodingType = ENCODING_NRZ;
				break;
			}
			
			case 1:
			{
				opp_error("\n[Wireless Channel]: Illegal value of parameter \"encodingType\" in omnetpp.ini. ENCODING_4B5B is not supported yet.\n" );
				break;
			}
			
			case 2:
			{
				opp_error("\n[Wireless Channel]: Illegal value of parameter \"encodingType\" in omnetpp.ini. ENCODING_MANCHESTER is not supported yet.\n" );
				break;
			}
			
			case 3:
			{
				opp_error("\n[Wireless Channel]: Illegal value of parameter \"encodingType\" in omnetpp.ini. ENCODING_SECDEC is not supported yet.\n" );
				break;
			}
			
			default:
			{
				opp_error("\n[Wireless Channel]: Illegal value of parameter \"encodingType\" in omnetpp.ini.\n" );
				break;
			}
		}
		
		// extract the txPowerLevels from the string module parameter
		const char *str = node_0->submodule("networkInterface")->submodule("Radio")->par("txPowerLevels");
		cStringTokenizer tokenizer(str);
		const char *token;
		numTxPowerLevels = 0;
		while ((token = tokenizer.nextToken())!=NULL){
			txPowerLevels[numTxPowerLevels] = atof(token);   // convert and store
			numTxPowerLevels++;
		}
		
		defaultPowerLevel = node_0->submodule("networkInterface")->submodule("Radio")->par("txPowerLevelUsed");
	}
	else
		opp_error("\n[Wireless Channel]:\n Error in geting a valid reference to  node[0] Application OR NetworkInterface (MAC/Radio) module.\n");
	
	
	
	collisionModel = par("collisionModel");
	switch(collisionModel)
	{
		case 0:
		{
			collisionModel = NO_INTERFERENCE_NO_COLLISIONS; 
			break;
		}
		
		case 1:
		{
			collisionModel = SIMPLE_COLLISION_MODEL; 
			break;
		}
		
		case 2:
		{
			collisionModel = ADDITIVE_INTERFERENCE_MODEL;
			break;
		}
		
		case 3:
		{
			collisionModel = COMPLEX_INTERFERENCE_MODEL;
			break;
		}
		
		default:
		{
			opp_error("\n[Wireless Channel]: Illegal value of parameter \"collisionModel\" in omnetpp.ini\n" );
			break;
		}
	}
	
	allBidirectionalLinks = par("allBidirectionalLinks");
}


void WChannel_Interference::parseRxPowerMap(void)
{
	cStringTokenizer valuesTokenizer(rxSignal_ConnectivityMap, ",");
	
	double * fields = new double [5];
	
	while (valuesTokenizer.hasMoreTokens()) 
	{
		const char * next = valuesTokenizer.nextToken();
		
		if(next == NULL)		
			continue;
		
		string token(next);
		
	
		if(token.size() <= 2)
			continue;
		
		string::size_type posA = token.find("(", 0);
		string::size_type posB = token.find(")", 0);
		string::size_type posC = token.find(">", 0); 
		string::size_type posD = token.find("=", 0);  
		
			 
		if( (posC == string::npos) || (posD == string::npos) || (( posA != string::npos) && (posB == string::npos)) )
			opp_error("\nError!(A) Illegal parameter format \"rxPowerMap\" of WChannel_Interference module.\n");
		
		
		cStringTokenizer fieldsTokenizer(token.c_str(), ">()=");
		
		
		int loop = 0;
		
		
		while(fieldsTokenizer.hasMoreTokens())
		{				
			string field(fieldsTokenizer.nextToken());
		
			string::size_type fieldSize = field.length();
			
			if(fieldSize >= 1)
			{
				fields[loop] = atof(field.c_str());
				loop++;
			}
			
		}
			
		if( ( posA != string::npos) && (loop < 4))
			opp_error("\nError!(B) Illegal parameter format \"rxPowerMap\" of WChannel_Interference module.\n");
		
		if( (fields[0] < 0) || (fields[0] >= numNodes) || (fields[1] < 0) || (fields[1] >= numNodes) )
			opp_error("\nError!(C) Illegal parameter format \"rxPowerMap\" of WChannel_Interference module.\n");
		
		if(loop == 3)
		{
			for (int k = 0; k < numTxPowerLevels; k++) 
				rxSignal[(int)fields[0]][(int)fields[1]][k] = fields[2] - txPowerLevels[k];
		}
		else
		if(loop == 4)
		{
			if( (fields[2] >= numTxPowerLevels) || (fields[2] < 0))
				opp_error("\nError!(D) Illegal parameter format \"rxPowerMap\" of WChannel_Interference module.\n");
				
			rxSignal[(int)fields[0]][(int)fields[1]][(int)fields[2]] = fields[3] - txPowerLevels[defaultPowerLevel];
		}
		else
		{
			opp_error("\nError!(E) Illegal parameter format \"rxPowerMap\" of WChannel_Interference module.\n");
		}

	}
	
	delete [] fields;
}


void WChannel_Interference::parsePrrMap(void)
{
	cStringTokenizer valuesTokenizer(PRR_ConnectivityMap, ",");
	
	double * fields = new double [5];
	
	double meanPacketSize = minumumPacketSize;
	
	bool doneOnce = false;
	
	while (valuesTokenizer.hasMoreTokens()) 
	{
		const char * next = valuesTokenizer.nextToken();
		
		if(next == NULL)		
			continue;
		
		string token(next);
	
		if(token.size() <= 2)
			continue;
		
		
		string::size_type posA = token.find("(", 0);
		string::size_type posB = token.find(")", 0);
		string::size_type posC = token.find(">", 0); 
		string::size_type posD = token.find("=", 0);
		string::size_type posE = token.find("<", 0);  
		
		
		if( posE != string::npos)
		{
			meanPacketSize = atof((token.substr(1, token.size()-1)).c_str());
			continue;
		}
		
		
		if( (posC == string::npos) || (posD == string::npos) || (( posA != string::npos) && (posB == string::npos)) )
			opp_error("\nError!(A) Illegal parameter format \"rxPowerMap\" of WChannel_Interference module.\n");
			
		
		cStringTokenizer fieldsTokenizer(token.c_str(), ">()=");
		
		
		int loop = 0;
		
		
		while(fieldsTokenizer.hasMoreTokens())
		{				
			string field(fieldsTokenizer.nextToken());
		
			string::size_type fieldSize = field.length();
			
			if(fieldSize >= 1)
			{
				fields[loop] = atof(field.c_str());
				//printf("%f # ", fields[loop]);
				loop++;
			}
			
		}
		//printf("\n");
			
		if( ( posA != string::npos) && (loop < 4))
			opp_error("\nError!(B) Illegal parameter format \"rxPowerMap\" of WChannel_Interference module.\n");
		
		if( (fields[0] < 0) || (fields[0] >= numNodes) || (fields[1] < 0) || (fields[1] >= numNodes) )
			opp_error("\nError!(C) Illegal parameter format \"rxPowerMap\" of WChannel_Interference module.\n");
		
		if(loop == 3)
		{
			double tmpSNRdB;
			if(modulationType == MODULATION_FSK)
				tmpSNRdB = 10.0 * log10(   (-2.0*(dataRate/noiseBandwidth)) * log( 2.0 * (1.0 -  pow(fields[2], 1.0/(8.0*meanPacketSize))) )   );
			else // (modulationType == MODULATION_PSK)
				tmpSNRdB = 10.0 * log10(   (dataRate/noiseBandwidth) * pow(erfcInv(2.0 * ( 1.0 - pow(fields[2], 1.0/(8.0*meanPacketSize)) )), 2.0)   );
			
			
			if(!doneOnce)
			{
				for (int i = 0; i < numNodes; i++)
					for (int j = 0; j < numNodes; j++)
							for (int k = 0; k < numTxPowerLevels; k++)
								rxSignal[i][j][k] = -100000;
				doneOnce = true;
			}
			
			for (int k = 0; k < numTxPowerLevels; k++) 
				rxSignal[(int)fields[0]][(int)fields[1]][k] = tmpSNRdB + noiseFloor;
		}
		else
		if(loop == 4)
		{
			double tmpSNRdB;
			if(modulationType == MODULATION_FSK)
				tmpSNRdB = 10.0 * log10(   (-2.0*(dataRate/noiseBandwidth)) * log( 2.0 * (1.0 -  pow(fields[3], 1.0/(8.0*meanPacketSize))) )   );
			else // (modulationType == MODULATION_PSK)
				tmpSNRdB = 10.0 * log10(   (dataRate/noiseBandwidth) * pow(erfcInv(2.0 * ( 1.0 - pow(fields[3], 1.0/(8.0*meanPacketSize)) )), 2.0)   );
			
			if( (fields[2] >= numTxPowerLevels) || (fields[2] < 0))
				opp_error("\nError!(D) Illegal parameter format \"rxPowerMap\" of WChannel_Interference module.\n");
				
			rxSignal[(int)fields[0]][(int)fields[1]][(int)fields[2]] = tmpSNRdB + noiseFloor;
		}
		else
		{
			opp_error("\nError!(E) Illegal parameter format \"rxPowerMap\" of WChannel_Interference module.\n");
		}

	}
	
	delete [] fields;
}


void WChannel_Interference::printRxSignalTable(void)
{
	for (int i = 0; i < numNodes; i++)
	{	
		for (int j = 0; j < numNodes; j++)
		{		
			printf("  [");
			for (int k = 0; k < numTxPowerLevels; k++)
			{
				printf("%f ", rxSignal[i][j][k]);
			}
			printf("]  ");
		}
		printf("\n");
	}
	printf("\n\n\n");
}


